try:
    from . import generic as g
except BaseException:
    import generic as g


def test_discrete():
    for d in g.get_2D():
        # store hash before requesting passive functions
        hash_val = d.__hash__()

        g.check_path2D(d)

        # make sure various methods return
        # basically the same bounds
        atol = d.scale / 1000
        for dis, pa, pl in zip(d.discrete, d.paths, d.polygons_closed):
            # bounds of discrete version of path
            bd = g.np.array([g.np.min(dis, axis=0), g.np.max(dis, axis=0)])
            # bounds of polygon version of path
            bl = g.np.reshape(pl.bounds, (2, 2))
            # try bounds of included entities from path
            pad = g.np.vstack([d.entities[i].discrete(d.vertices) for i in pa])
            bp = g.np.array([g.np.min(pad, axis=0), g.np.max(pad, axis=0)])

            assert g.np.allclose(bd, bl, atol=atol)
            assert g.np.allclose(bl, bp, atol=atol)

        # run some checks
        g.check_path2D(d)

        # copying shouldn't touch original file
        copied = d.copy()

        # these operations shouldn't have mutated anything!
        assert d.__hash__() == hash_val

        # copy should have saved the metadata
        assert set(copied.metadata.keys()) == set(d.metadata.keys())

        # file_name should be populated, and if we have a DXF file
        # the layer field should be populated with layer names
        if d.source.file_name[-3:] == "dxf":
            assert len(d.layers) == len(d.entities)

        for path, verts in zip(d.paths, d.discrete):
            assert len(path) >= 1
            dists = g.np.sum((g.np.diff(verts, axis=0)) ** 2, axis=1) ** 0.5

            if not g.np.all(dists > g.tol_path.zero):
                raise ValueError("{} had zero distance in discrete!", d.source.file_name)

            circuit_dist = g.np.linalg.norm(verts[0] - verts[-1])
            circuit_test = circuit_dist < g.tol_path.merge
            if not circuit_test:
                g.log.error(
                    "On file %s First and last vertex distance %f",
                    d.source.file_name,
                    circuit_dist,
                )
            assert circuit_test

            is_ccw = g.trimesh.path.util.is_ccw(verts)
            if not is_ccw:
                g.log.error("discrete %s not ccw!", d.source.file_name)

        for i in range(len(d.paths)):
            assert d.polygons_closed[i].is_valid
            assert d.polygons_closed[i].area > g.tol_path.zero
        export_dict = d.export(file_type="dict")
        to_dict = d.to_dict()
        assert isinstance(to_dict, dict)
        assert isinstance(export_dict, dict)
        assert len(to_dict) == len(export_dict)

        export_svg = d.export(file_type="svg")  # NOQA
        simple = d.simplify()  # NOQA
        split = d.split()
        g.log.info(
            "Split %s into %d bodies, checking identifiers",
            d.source.file_name,
            len(split),
        )
        for body in split:
            _ = body.identifier

        if len(d.root) == 1:
            d.apply_obb()

        # store the X values of bounds
        ori = d.bounds.copy()
        # apply a translation
        d.apply_translation([10, 0])
        # X should have translated by 10.0
        assert g.np.allclose(d.bounds[:, 0] - 10, ori[:, 0])
        # Y should not have moved
        assert g.np.allclose(d.bounds[:, 1], ori[:, 1])

        if len(d.polygons_full) > 0 and len(d.vertices) < 150:
            g.log.info("Checking medial axis on %s", d.source.file_name)
            m = d.medial_axis()
            assert len(m.entities) > 0

        # shouldn't crash
        d.fill_gaps()

        # transform to first quadrant
        d.rezero()
        # run process manually
        d.process()


def test_path():
    path = g.trimesh.path.Path3D(
        entities=[g.trimesh.path.entities.Line(points=[0, 1, 2])],
        vertices=[[0, 0, 0], [1, 0, 0], [1, 1, 0]],
        vertex_attributes={"_test": g.np.array([1, 2, 3], dtype=g.np.float32)},
    )

    glb_data = g.to_glb_bytes(path)
    loaded_scene = g.from_glb_bytes(glb_data)

    loaded_path = loaded_scene.geometry["geometry_0"]
    loaded_path.process()

    assert isinstance(loaded_path, g.trimesh.path.Path3D)

    assert loaded_path.vertices.shape == (3, 3)
    assert len(loaded_path.entities) == 1
    assert len(loaded_path.vertex_attributes["_test"]) == 3


def test_path_to_gltf_with_line():
    path = g.trimesh.path.Path3D(
        entities=[
            g.trimesh.path.entities.Line(points=[0, 1, 2]),
            g.trimesh.path.entities.Line(points=[3, 4, 5]),
        ],
        vertices=g.np.array(
            [
                [0, 0, 0],
                [0, 1, 0],
                [1, 1, 0],
                [0, 0, 1],
                [0, 1, 1],
                [1, 1, 2],
            ]
        ),
    )
    path.vertex_attributes = {"_test": g.np.array([0, 0, 0, 1, 1, 1], dtype=g.np.float32)}

    tree, _ = g.trimesh.exchange.gltf._create_gltf_structure(g.trimesh.Scene([path]))
    # should have included the attribute by name
    assert set(tree["meshes"][0]["primitives"][0]["attributes"].keys()) == {
        "POSITION",
        "_test",
    }

    assert len(tree["accessors"]) == 2
    acc = tree["accessors"]
    assert acc[0]["count"] == acc[1]["count"]


def test_path_to_gltf_with_arc():
    path = g.trimesh.path.Path3D(
        entities=[
            g.trimesh.path.entities.Line(points=[0, 1, 2]),
            g.trimesh.path.entities.Arc(points=[3, 4, 5]),
        ],
        vertices=g.np.array(
            [
                [0, 0, 0],
                [0, 1, 0],
                [1, 1, 0],
                [0, 0, 1],
                [0, 1, 1],
                [1, 1, 2],
            ]
        ),
    )
    path.vertex_attributes = {"_test": g.np.array([0, 0, 0, 1, 1, 1], dtype=g.np.float32)}

    tree, _ = g.trimesh.exchange.gltf._create_gltf_structure(g.trimesh.Scene([path]))

    # should have skipped the attribute rather than exporting a broken one
    assert set(tree["meshes"][0]["primitives"][0]["attributes"].keys()) == {"POSITION"}


def test_poly():
    p = g.get_mesh("2D/LM2.dxf")
    assert p.is_closed

    # one of the lines should be a polyline
    assert any(
        len(e.points) > 2
        for e in p.entities
        if isinstance(e, g.trimesh.path.entities.Line)
    )

    # layers should match entity count
    assert len(p.layers) == len(p.entities)
    assert len(set(p.layers)) > 1

    # check copy with layer specified
    unique = set(p.layers)
    assert len(unique) > 1
    for u in unique:
        a = p.copy(layers={u})
        assert len(a.entities) > 0
        assert p.layers.count(u) == len(a.entities)
        assert len(a.entities) < len(p.layers)
        assert all(e.layer == u for e in a.entities)

    # try exploding
    count = len(p.entities)

    p.explode()
    # explode should have created new entities
    assert len(p.entities) > count
    # explode should have added some new layers
    assert len(p.entities) == len(p.layers)
    # all line segments should have two points now
    assert all(
        len(i.points) == 2
        for i in p.entities
        if isinstance(i, g.trimesh.path.entities.Line)
    )
    # should still be closed
    assert p.is_closed
    # chop off the last entity
    p.entities = p.entities[:-1]
    # should no longer be closed
    assert not p.is_closed

    # fill gaps of any distance
    p.fill_gaps(g.np.inf)
    # should have fixed this puppy
    assert p.is_closed

    # remove 2 short edges using remove_entity()
    count = len(p.entities)
    p.remove_entities([count - 1, count - 6])
    assert not p.is_closed
    p.fill_gaps(2)
    assert p.is_closed


def test_text():
    """
    Do some checks on Text entities
    """
    p = g.get_mesh("2D/LM2.dxf")
    p.explode()
    # get some text entities
    text = [e for e in p.entities if isinstance(e, g.trimesh.path.entities.Text)]
    assert len(text) > 1

    # loop through each of them
    for t in text:
        # a spurious error we were seeing in CI
        if g.trimesh.util.is_instance_named(t, "Line"):
            raise ValueError(
                "type bases:", [i.__name__ for i in g.trimesh.util.type_bases(t)]
            )
    # make sure this doesn't crash with text entities
    g.trimesh.rendering.convert_to_vertexlist(p)


def test_empty():
    # make sure empty paths perform as expected

    p = g.trimesh.path.Path2D()
    assert p.is_empty
    p = g.trimesh.path.Path3D()
    assert p.is_empty

    p = g.trimesh.load_path([[[0, 0, 1], [1, 0, 1]]])
    assert len(p.entities) == 1
    assert not p.is_empty

    b = p.to_planar()[0]
    assert len(b.entities) == 1
    assert not b.is_empty


def test_color():
    p = g.get_mesh("2D/wrench.dxf")
    # make sure we have entities
    assert len(p.entities) > 0
    # make sure shape of colors is correct
    assert all(e.color is None for e in p.entities)
    assert p.colors is None

    color = [255, 0, 0, 255]
    # assign a color to the entity
    p.entities[0].color = color
    # make sure this is reflected in the path color
    assert g.np.allclose(p.colors[0], color)
    assert p.colors.shape == (len(p.entities), 4)

    p.colors = g.np.array([100, 100, 100] * len(p.entities), dtype=g.np.uint8).reshape(
        (-1, 3)
    )
    assert g.np.allclose(p.colors[0], [100, 100, 100, 255])


def test_dangling():
    p = g.get_mesh("2D/wrench.dxf")
    assert len(p.dangling) == 0

    b = g.get_mesh("2D/loose.dxf")
    assert len(b.dangling) == 1
    assert len(b.polygons_full) == 1


def test_plot():
    try:
        # only run these if matplotlib is installed
        import matplotlib.pyplot  # NOQA
    except BaseException:
        g.log.debug("skipping `matplotlib.pyplot` tests")
        return

    p = g.get_mesh("2D/wrench.dxf")

    # see if the logic crashes
    p.plot_entities(show=False)
    p.plot_discrete(show=False)


def test_split():
    for fn in [
        "2D/ChuteHolderPrint.DXF",
        "2D/tray-easy1.dxf",
        "2D/sliding-base.dxf",
        "2D/wrench.dxf",
        "2D/spline_1.dxf",
    ]:
        p = g.get_mesh(fn)

        # make sure something was loaded
        assert len(p.root) > 0

        # split by connected
        split = p.split()

        # make sure split parts have same area as source
        assert g.np.isclose(p.area, sum(i.area for i in split))
        # make sure concatenation doesn't break that
        assert g.np.isclose(p.area, g.np.sum(split).area)

        # check that cache didn't screw things up
        for s in split:
            assert len(s.root) == 1
            assert len(s.path_valid) == len(s.paths)
            assert len(s.paths) == len(s.discrete)
            assert s.path_valid.sum() == len(s.polygons_closed)
            g.check_path2D(s)


def test_section():
    mesh = g.get_mesh("tube.obj")

    # check the CCW correctness with a normal in both directions
    for sign in [1.0, -1.0]:
        # get a cross section of the tube
        section = mesh.section(
            plane_origin=mesh.center_mass, plane_normal=[0.0, sign, 0.0]
        )

        # Path3D -> Path2D
        planar, _T = section.to_planar()

        # tube should have one closed polygon
        assert len(planar.polygons_full) == 1
        polygon = planar.polygons_full[0]
        # closed polygon should have one interior
        assert len(polygon.interiors) == 1

        # the exterior SHOULD be counterclockwise
        assert g.trimesh.path.util.is_ccw(polygon.exterior.coords)
        # the interior should NOT be counterclockwise
        assert not g.trimesh.path.util.is_ccw(polygon.interiors[0].coords)

        # should be a valid Path2D
        g.check_path2D(planar)


def test_multiplane():
    # check to make sure we're applying `tol_path.merge_digits` for line segments
    vertices = g.np.array(
        [
            [19.402250931139097, -14.88787016674277, 64.0],
            [20.03318396099334, -14.02738654377374, 64.0],
            [19.402250931139097, -14.88787016674277, 0.0],
            [20.03318396099334, -14.02738654377374, 0.0],
            [21, -16.5, 32],
        ]
    )
    faces = g.np.array([[1, 3, 0], [0, 3, 2], [1, 0, 4], [0, 2, 4], [3, 1, 4], [2, 3, 4]])
    z_layer_centers = g.np.array([3, 3.0283334255218506])
    m = g.trimesh.Trimesh(vertices, faces)
    r = m.section_multiplane([0, 0, 0], [0, 0, 1], z_layer_centers)
    assert len(r) == 2
    assert all(i.is_closed for i in r)


def test_svg_arc_snap():
    s = """<svg xmlns="http://www.w3.org/2000/svg" width="97.5mm" height="35mm" viewBox="-17.5 0 97.5 35">

<path d="M 0 0 h 80 v 35 h -80 a 17.5 17.5 0 0 1 0 -35 z" fill="none" stroke="black" stroke-width="0.1" />
</svg>"""

    p = g.trimesh.load_path(g.trimesh.util.wrap_as_stream(s), file_type="svg")

    # should be one closed path
    assert len(p.root) == 1
    # should be exactly closed with a snap
    assert g.np.equal(*p.discrete[0][[0, -1]]).all()
    # the extrusion should be a volume
    assert p.extrude(10.0).is_volume


def test_path_dict_export():
    x, y = 2, 3

    # define incorrectly
    as_dict = {
        "entities": [
            {"type": "Line", "points": [0, 1, 2, 3, 0], "closed": False},
        ],
        "vertices": [
            [-x / 2, y / 2],
            [x / 2, y / 2],
            [x / 2, -y / 2],
            [-x / 2, -y / 2],
        ],
    }

    # try loading with the incorrect `closed` value for the line
    path = g.trimesh.path.exchange.load.load_path(
        g.trimesh.path.exchange.misc.dict_to_path(as_dict)
    )

    # export to a dict again
    exp = path.export(file_type="dict")
    assert len(exp["entities"]) == 1
    assert "closed" not in exp["entities"][0]

    # do a roundtrip
    rt = g.trimesh.path.exchange.load.load_path(
        g.trimesh.path.exchange.misc.dict_to_path(exp)
    )

    # should have the same area
    assert g.np.isclose(rt.area, path.area)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    # g.unittest.main()

    # test_path_to_gltf_with_arc()
    # test_path_dict_export()
    test_poly()



from math import log2

from .typed import Any, Callable, Iterable, List, NDArray, Sequence, Union


def reduce_cascade(operation: Callable, items: Union[Sequence, NDArray]):
    """
    Call an operation function in a cascaded pairwise way against a
    flat list of items.

    This should produce the same result as `functools.reduce`
    if `operation` is commutable like addition or multiplication.
    This may be faster for an `operation` that runs with a speed
    proportional to its largest input, which mesh booleans appear to.

    The union of a large number of small meshes appears to be
    "much faster" using this method.

    This only differs from `functools.reduce` for commutative `operation`
    in that it returns `None` on empty inputs rather than `functools.reduce`
    which raises a `TypeError`.

    For example on `a b c d e f g` this function would run and return:
        a b
        c d
        e f
        ab cd
        ef g
        abcd efg
     -> abcdefg

    Where `functools.reduce` would run and return:
        a b
        ab c
        abc d
        abcd e
        abcde f
        abcdef g
     -> abcdefg

    Parameters
    ----------
    operation
      The function to call on pairs of items.
    items
      The flat list of items to apply operation against.
    """
    if len(items) == 0:
        return None
    elif len(items) == 1:
        # skip the loop overhead for a single item
        return items[0]
    elif len(items) == 2:
        # skip the loop overhead for a single pair
        return operation(items[0], items[1])

    for _ in range(int(1 + log2(len(items)))):
        results = []

        # loop over pairs of items.
        items_mod = len(items) % 2
        for i in range(0, len(items) - items_mod, 2):
            results.append(operation(items[i], items[i + 1]))

        # if we had a non-even number of items it will have been
        # skipped by the loop so append it to our list
        if items_mod != 0:
            results.append(items[-1])

        items = results

    # logic should have reduced to a single item
    assert len(results) == 1

    return results[0]


def chain(*args: Union[Iterable[Any], Any, None]) -> List[Any]:
    """
    A less principled version of `list(itertools.chain(*args))` that
    accepts non-iterable values, filters `None`, and returns a list
    rather than yielding values.

    If all passed values are iterables this will return identical
    results to `list(itertools.chain(*args))`.


    Examples
    ----------

    In [1]: list(itertools.chain([1,2], [3]))
    Out[1]: [1, 2, 3]

    In [2]: trimesh.util.chain([1,2], [3])
    Out[2]: [1, 2, 3]

    In [3]: trimesh.util.chain([1,2], [3], 4)
    Out[3]: [1, 2, 3, 4]

    In [4]: list(itertools.chain([1,2], [3], 4))
      ----> 1 list(itertools.chain([1,2], [3], 4))
      TypeError: 'int' object is not iterable

    In [5]: trimesh.util.chain([1,2], None, 3, None, [4], [], [], 5, [])
    Out[5]: [1, 2, 3, 4, 5]


    Parameters
    -----------
    args
      Will be individually checked to see if they're iterable
      before either being appended or extended to a flat list.


    Returns
    ----------
    chained
      The values in a flat list.
    """
    # collect values to a flat list
    chained = []
    # extend if it's a sequence, otherwise append
    [
        chained.extend(a)
        if (hasattr(a, "__iter__") and not isinstance(a, (str, bytes)))
        else chained.append(a)
        for a in args
        if a is not None
    ]
    return chained



try:
    from . import generic as g
except BaseException:
    import generic as g


class IntervalTest(g.unittest.TestCase):
    def test_intersection(self):
        pairs = g.np.array(
            [
                [[0, 1], [1, 2]],
                [[1, 0], [1, 2]],
                [[0, 0], [0, 0]],
                [[10, 20], [9, 21]],
                [[5, 15], [7, 10]],
                [[5, 10], [10, 9]],
                [[0, 1], [0.9, 10]],
                [[1000, 1001], [2000, 2001]],
            ],
            dtype=g.np.float64,
        )

        # true intersection ranges
        truth = g.np.array(
            [
                [0.0, 0.0],
                [0.0, 0.0],
                [0.0, 0.0],
                [10, 20],
                [7, 10],
                [9, 10],
                [0.9, 1.0],
                [0, 0],
            ],
            dtype=g.np.float64,
        )

        intersection = g.trimesh.interval.intersection
        union = g.trimesh.interval.union

        # check the single- interval results
        for ab, tru in zip(pairs, truth):
            result = intersection(*ab)
            assert g.np.allclose(result, tru)

        # check the vectorized multiple interval results
        inter = intersection(pairs[:, 0, :], pairs[:, 1, :])

        assert g.np.allclose(truth, inter)

        # now just run a union on these for the fun of it
        u = union(pairs.reshape((-1, 2)))
        assert g.np.allclose(u, [[0.0, 21.0], [1000.0, 1001.0], [2000.0, 2001.0]])


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_quadric_simplification():
    if not g.trimesh.util.has_module("fast_simplification"):
        return

    m = g.get_mesh("rabbit.obj")
    assert len(m.faces) > 600

    # should be about half as large
    a = m.simplify_quadric_decimation(percent=0.5)
    assert g.np.isclose(len(a.faces), len(m.faces) // 2, rtol=0.2)

    # should have the requested number of faces
    a = m.simplify_quadric_decimation(face_count=200)
    assert len(a.faces) == 200

    # see if aggression does anything
    a = m.simplify_quadric_decimation(percent=0.25, aggression=5)
    assert len(a.faces) > 0


if __name__ == "__main__":
    test_quadric_simplification()



try:
    from . import generic as g
except BaseException:
    import generic as g

import io


class ExportTest(g.unittest.TestCase):
    def test_export(self):
        from trimesh.exceptions import ExceptionWrapper

        export_types = {
            k
            for k, v in g.trimesh.exchange.export._mesh_exporters.items()
            if not isinstance(v, ExceptionWrapper)
        }

        meshes = list(g.get_meshes(10))
        # make sure we've got something with texture
        meshes.append(g.get_mesh("fuze.obj"))

        for mesh in meshes:
            # disregard texture
            mesh.merge_vertices(merge_tex=True, merge_norm=True)
            for file_type in export_types:
                # skip pointcloud format
                if file_type in ["xyz", "gltf"]:
                    # a pointcloud format
                    continue
                # run the export
                export = mesh.export(file_type=file_type)
                # if nothing returned log the message
                if export is None or len(export) == 0:
                    raise ValueError(
                        "No data exported %s to %s", mesh.source.file_name, file_type
                    )

                if mesh.visual.kind == "texture":
                    c = mesh.copy()
                    c.visual.uv = None
                    c.export(file_type=file_type)

                if file_type in [
                    "dae",  # collada, no native importers
                    "collada",  # collada, no native importers
                    "msgpack",  # kind of flaky, but usually works
                    "drc",
                ]:  # DRC is not a lossless format
                    g.log.warning("no native loaders implemented for collada!")
                    continue

                g.log.info("Export/import testing on %s", mesh.source.file_name)

                if isinstance(export, str):
                    assert export.endswith("\n"), f"{file_type} doesn't end with newline"

                # if export is string or bytes wrap as pseudo file object
                if isinstance(export, str) or isinstance(export, bytes):
                    file_obj = g.io_wrap(export)
                else:
                    file_obj = export

                loaded = g.trimesh.load(
                    file_obj=file_obj,
                    file_type=file_type,
                    process=True,
                    merge_norm=True,
                    merge_tex=True,
                )

                # if we exported as GLTF/dae it will come back as a Scene
                if isinstance(loaded, g.trimesh.Scene) and isinstance(
                    mesh, g.trimesh.Trimesh
                ):
                    assert len(loaded.geometry) == 1
                    loaded = next(iter(loaded.geometry.values()))

                if (
                    not g.trimesh.util.is_shape(loaded._data["faces"], (-1, 3))
                    or not g.trimesh.util.is_shape(loaded._data["vertices"], (-1, 3))
                    or loaded.faces.shape != mesh.faces.shape
                ):
                    g.log.error(
                        "Export -> import for %s on %s wrong shape!",
                        file_type,
                        mesh.source.file_name,
                    )

                if loaded.vertices is None:
                    g.log.error(
                        "Export -> import for %s on %s gave None for vertices!",
                        file_type,
                        mesh.source.file_name,
                    )

                if loaded.faces.shape != mesh.faces.shape:
                    raise ValueError(
                        f"export cycle {file_type} on {mesh.source.file_name} gave faces {mesh.faces.shape!s}->{loaded.faces.shape!s}!"
                    )

                if loaded.vertices.shape != mesh.vertices.shape:
                    raise ValueError(
                        f"export cycle {file_type} on {mesh.source.file_name} gave vertices {mesh.vertices.shape}->{loaded.vertices.shape}!"
                    )

                # try exporting/importing certain file types by name
                if file_type in ["obj", "stl", "ply", "off"]:
                    temp = g.tempfile.NamedTemporaryFile(
                        suffix="." + file_type, delete=False
                    )
                    # windows throws permissions errors if you keep it open
                    temp.close()

                    mesh.export(temp.name)
                    load = g.trimesh.load(temp.name)
                    # manual cleanup
                    g.os.remove(temp.name)

                    assert mesh.faces.shape == load.faces.shape
                    assert mesh.vertices.shape == load.vertices.shape

            # if we're not on linux don't run meshlab tests
            if not g.is_linux:
                continue
            # formats exportable by trimesh and importable by meshlab
            # make sure things we export can be loaded by meshlab
            both = set(g.meshlab_formats).intersection(set(export_types))

            # additional options to pass to exporters to try to ferret
            # out combinations which lead to invalid output
            kwargs = {
                "ply": [
                    {"vertex_normal": True, "encoding": "ascii"},
                    {"vertex_normal": True, "encoding": "binary"},
                    {"vertex_normal": False, "encoding": "ascii"},
                    {"vertex_normal": False, "encoding": "binary"},
                ],
                "stl": [{"file_type": "stl"}, {"file_type": "stl_ascii"}],
            }

            # make sure input mesh has garbage removed
            mesh._validate = True
            # since we're going to be looking for exact export
            # counts remove anything small/degenerate again
            mesh.process()

            # run through file types supported by both meshlab and trimesh
            for file_type in both:
                # pull different exporter options for the format
                if file_type in kwargs:
                    options = kwargs[file_type]
                else:
                    options = [{}]

                # try each combination of options
                for option in options:
                    temp = g.tempfile.NamedTemporaryFile(
                        suffix="." + file_type, delete=False
                    )
                    temp_off = g.tempfile.NamedTemporaryFile(suffix=".off", delete=False)
                    # windows throws permissions errors if you keep it open
                    temp.close()
                    temp_off.close()
                    # write over the tempfile
                    option["file_obj"] = temp.name
                    mesh.export(**option)

                    # -_-
                    ms = g.pymeshlab.MeshSet()
                    ms.load_new_mesh(temp.name)
                    ms.save_current_mesh(temp_off.name)

                    # load meshlabs export back into trimesh
                    r = g.trimesh.load(temp_off.name)

                    # we should have the same number of vertices and faces
                    assert len(r.vertices) == len(mesh.vertices)
                    assert len(r.faces) == len(mesh.faces)

                    if option.get("encoding", None) == "ascii":
                        with open(temp.name) as f:
                            exported = f.read()
                        assert exported.endswith("\n")

                    # manual cleanup
                    g.os.remove(temp.name)
                    g.os.remove(temp_off.name)

    def test_obj(self):
        m = g.get_mesh("textured_tetrahedron.obj", process=False)
        export = m.export(file_type="obj")
        reconstructed = g.roundtrip(export, file_type="obj", process=False)
        # test that we get at least the same number of normals and texcoords out;
        # the loader may reorder vertices, so we shouldn't check direct
        # equality
        assert m.vertex_normals.shape == reconstructed.vertex_normals.shape

    def test_obj_order(self):
        """
        Make sure simple round trips through Wavefront don't
        reorder vertices.
        """
        # get a writeable temp file location
        temp = g.tempfile.NamedTemporaryFile(suffix=".obj", delete=False)
        temp.close()

        # simple solid
        x = g.trimesh.creation.icosahedron()
        x.export(temp.name)
        y = g.trimesh.load_mesh(temp.name, process=False)

        # vertices should be same shape and order
        assert g.np.allclose(x.vertices, y.vertices)
        # faces should be same
        assert g.np.allclose(x.faces, y.faces)

    def test_dict(self):
        mesh = g.get_mesh("machinist.XAML")
        assert mesh.visual.kind == "face"
        mesh.visual.vertex_colors = mesh.visual.vertex_colors
        assert mesh.visual.kind == "vertex"

        as_dict = mesh.to_dict()
        back = g.trimesh.Trimesh(**as_dict, process=False)
        assert g.np.allclose(back.vertices, mesh.vertices)

    def test_scene(self):
        # get a multi- mesh scene with a transform tree
        source = g.get_mesh("cycloidal.3DXML")
        # add a transform to zero scene before exporting
        source.rezero()
        # export the file as a binary GLTF file, GLB
        export = source.export(file_type="glb")

        # re- load the file as a trimesh.Scene object again
        loaded = g.roundtrip(export, file_type="glb")

        # the scene should be identical after export-> import cycle
        assert g.np.allclose(loaded.extents / source.extents, 1.0)

    def test_ply_path_empty(self):
        """
        Test empty path export does not fail
        """
        path3D = g.trimesh.path.Path3D()

        # export to ply
        ply = path3D.export(file_type="ply")
        assert len(ply) > 0

        loaded = g.trimesh.load_path(g.trimesh.util.wrap_as_stream(ply), file_type="ply")
        # assert g.np.allclose(loaded.entities, path3D.entities)
        # assert g.np.allclose(loaded.vertices, path3D.vertices)
        assert len(loaded.geometry) == 0

    def test_ply_path_line(self):
        """
        Should be able to load a path and export simple line as a PLY
        """
        path3D = g.trimesh.load_path([(0, 0, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1)])
        assert isinstance(path3D, g.trimesh.path.Path3D)

        # export to ply
        ply = path3D.export(file_type="ply")
        assert len(ply) > 0

        loaded = g.trimesh.load_path(g.trimesh.util.wrap_as_stream(ply), file_type="ply")
        assert g.np.allclose(loaded.vertices, path3D.vertices)

    def test_ply_path_multi(self):
        """
        Should be able to load a path and export multiple entities as a PLY
        """
        path3D = g.trimesh.path.Path3D(
            [
                g.trimesh.path.entities.Line([0, 1, 2]),
                g.trimesh.path.entities.Line([2, 0, 1]),
            ],
            [(0.0, 0.0, 0.0), (1.0, 0.0, 0.0), (1.0, 1.0, 0.0)],
        )

        # export to ply
        ply = path3D.export(file_type="ply")
        assert len(ply) > 0

        loaded = g.trimesh.load_path(g.trimesh.util.wrap_as_stream(ply), file_type="ply")
        assert len(loaded.entities) == len(path3D.entities)
        assert len(loaded.vertices) > 0

    def test_ply_path_bezier(self):
        """
        Should be able to load a path and export complex curve as a PLY
        """
        path3D = g.trimesh.path.Path3D(
            [g.trimesh.path.entities.Bezier([0, 1, 2])],
            [(0.0, 0.0, 0.0), (1.0, 0.0, 0.0), (1.0, 1.0, 0.0)],
        )

        # export to ply
        ply = path3D.export(file_type="ply")
        assert len(ply) > 0

        loaded = g.trimesh.load_path(g.trimesh.util.wrap_as_stream(ply), file_type="ply")
        # note: we cannot recover the exact entities and vertices from the export since it is discretized for ply files
        assert len(loaded.entities) > 0
        assert len(loaded.vertices) > 0

    def test_gltf_path(self):
        """
        Check to make sure GLTF exports of Path2D and Path3D
        objects don't immediately crash.
        """
        path2D = g.get_mesh("2D/wrench.dxf")
        path3D = path2D.to_3D()

        a = g.trimesh.Scene(path2D).export(file_type="glb")
        b = g.trimesh.Scene(path3D).export(file_type="glb")

        assert len(a) > 0
        assert len(b) > 0

    def test_parse_file_args(self):
        """
        Test the magical trimesh.exchange.load.parse_file_args
        """
        # it's wordy
        f = g.trimesh.exchange.load._parse_file_args

        # a path that doesn't exist
        nonexists = f"/banana{g.random()}"
        assert not g.os.path.exists(nonexists)

        # loadable OBJ model
        exists = g.os.path.join(g.dir_models, "tube.obj")
        assert g.os.path.exists(exists)

        # should be able to extract type from passed filename
        args = f(file_obj=exists, file_type=None)
        assert args.file_type == "obj"

        # should be able to extract correct type from longer name
        args = f(file_obj=exists, file_type="YOYOMA.oBj")
        assert args.file_type == "obj"

        # with a nonexistent file and no extension it should raise
        try:
            args = f(file_obj=nonexists, file_type=None)
        except ValueError as E:
            assert "not a file" in str(E)
        else:
            raise ValueError("should have raised exception!")

        # nonexistent file with extension passed should return
        # file name anyway, maybe something else can handle it
        args = f(file_obj=nonexists, file_type=".ObJ")
        # should have cleaned up case
        assert args.file_type == "obj"

        # make sure overriding type works for string filenames
        args = f(file_obj=exists, file_type="STL")
        # should have used manually passed type over .obj
        assert args.file_type == "stl"

    def test_buffered_random(self):
        """Test writing to non-standard file"""
        mesh = next(iter(g.get_meshes(1)))
        with io.BufferedRandom(io.BytesIO()) as rw:
            mesh.export(rw, "STL")
            rw.seek(0)
            binary_stl = rw.read()
            self.assertLess(0, len(binary_stl))


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



import json
import os
from copy import deepcopy

import numpy as np

from .. import resolvers, util
from ..base import Trimesh
from ..exceptions import ExceptionWrapper
from ..parent import Geometry, LoadSource
from ..points import PointCloud
from ..scene.scene import Scene, append_scenes
from ..typed import Dict, Loadable, Optional, Set
from ..util import log
from . import misc
from .binvox import _binvox_loaders
from .cascade import _cascade_loaders
from .dae import _collada_loaders
from .gltf import _gltf_loaders
from .misc import _misc_loaders
from .obj import _obj_loaders
from .off import _off_loaders
from .ply import _ply_loaders
from .stl import _stl_loaders
from .threedxml import _threedxml_loaders
from .threemf import _three_loaders
from .xaml import _xaml_loaders
from .xyz import _xyz_loaders

try:
    from ..path.exchange.load import load_path, path_formats
except BaseException as E:
    # save a traceback to see why path didn't import
    load_path = ExceptionWrapper(E)

    # no path formats available
    def path_formats() -> set:
        return set()


def mesh_formats() -> Set[str]:
    """
    Get a list of mesh formats available to load.

    Returns
    -----------
    loaders
      Extensions of available mesh loaders
      i.e. `{'stl', 'ply'}`
    """
    # filter out exceptionmodule loaders
    return {k for k, v in mesh_loaders.items() if not isinstance(v, ExceptionWrapper)}


def available_formats() -> Set[str]:
    """
    Get a list of all available loaders


    Returns
    -----------
    loaders
      Extensions of all available loaders
      i.e. `{'stl', 'ply', 'dxf'}`
    """
    loaders = mesh_formats()
    loaders.update(path_formats())
    loaders.update(compressed_loaders.keys())

    return loaders


def load(
    file_obj: Loadable,
    file_type: Optional[str] = None,
    resolver: Optional[resolvers.ResolverLike] = None,
    force: Optional[str] = None,
    allow_remote: bool = False,
    **kwargs,
) -> Geometry:
    """
    THIS FUNCTION IS DEPRECATED but there are no current plans for it to be removed.

    For new code the typed load functions `trimesh.load_scene` or `trimesh.load_mesh`
    are recommended over `trimesh.load` which is a backwards-compatibility wrapper
    that mimics the behavior of the old function and can return any geometry type.

    Parameters
    -----------
    file_obj : str, or file- like object
      The source of the data to be loadeded
    file_type: str
      What kind of file type do we have (eg: 'stl')
    resolver : trimesh.visual.Resolver
      Object to load referenced assets like materials and textures
    force : None or str
      For 'mesh': try to coerce scenes into a single mesh
      For 'scene': try to coerce everything into a scene
    allow_remote
      If True allow this load call to work on a remote URL.
    kwargs : dict
      Passed to geometry __init__

    Returns
    ---------
    geometry : Trimesh, Path2D, Path3D, Scene
      Loaded geometry as trimesh classes
    """

    # call the most general loading case into a `Scene`.
    loaded = load_scene(
        file_obj=file_obj,
        file_type=file_type,
        resolver=resolver,
        allow_remote=allow_remote,
        **kwargs,
    )

    if force == "mesh":
        # new code should use `load_mesh` for this
        log.debug(
            "`trimesh.load(force='mesh')` is a compatibility wrapper for `trimesh.load_mesh`"
        )
        return loaded.to_mesh()
    elif force == "scene":
        # new code should use `load_scene` for this
        log.debug(
            "`trimesh.load(force='scene')` is a compatibility wrapper for `trimesh.load_scene`"
        )
        return loaded

    ###########################################
    # we are matching old, deprecated behavior here!
    kind = loaded.source.file_type
    always_scene = {"glb", "gltf", "zip", "3dxml", "tar.gz"}

    if kind not in always_scene and len(loaded.geometry) == 1:
        geom = next(iter(loaded.geometry.values()))
        geom.metadata.update(loaded.metadata)

        if isinstance(geom, PointCloud) or kind in {
            "obj",
            "stl",
            "ply",
            "svg",
            "binvox",
            "xaml",
            "dxf",
            "off",
            "msh",
        }:
            return geom

    return loaded


def load_scene(
    file_obj: Loadable,
    file_type: Optional[str] = None,
    resolver: Optional[resolvers.ResolverLike] = None,
    allow_remote: bool = False,
    metadata: Optional[Dict] = None,
    **kwargs,
) -> Scene:
    """
    Load geometry into the `trimesh.Scene` container. This may contain
    any `parent.Geometry` object, including `Trimesh`, `Path2D`, `Path3D`,
    or a `PointCloud`.

    Parameters
    -----------
    file_obj : str, or file- like object
      The source of the data to be loadeded
    file_type: str
      What kind of file type do we have (eg: 'stl')
    resolver : trimesh.visual.Resolver
      Object to load referenced assets like materials and textures
    force : None or str
      For 'mesh': try to coerce scenes into a single mesh
      For 'scene': try to coerce everything into a scene
    allow_remote
      If True allow this load call to work on a remote URL.
    kwargs : dict
      Passed to geometry __init__

    Returns
    ---------
    geometry : Trimesh, Path2D, Path3D, Scene
      Loaded geometry as trimesh classes
    """

    # parse all possible values of file objects into simple types
    arg = _parse_file_args(
        file_obj=file_obj,
        file_type=file_type,
        resolver=resolver,
        allow_remote=allow_remote,
    )

    try:
        if isinstance(file_obj, dict):
            # we've been passed a dictionary so treat them as keyword arguments
            loaded = _load_kwargs(file_obj)
        elif arg.file_type in path_formats():
            # use path loader
            loaded = load_path(
                file_obj=arg.file_obj,
                file_type=arg.file_type,
                metadata=metadata,
                **kwargs,
            )
        elif arg.file_type in mesh_loaders:
            # use mesh loader
            parsed = deepcopy(kwargs)
            parsed.update(
                mesh_loaders[arg.file_type](
                    file_obj=arg.file_obj,
                    file_type=arg.file_type,
                    resolver=arg.resolver,
                    metadata=metadata,
                    **kwargs,
                )
            )
            loaded = _load_kwargs(**parsed)

        elif arg.file_type in compressed_loaders:
            # for archives, like ZIP files
            loaded = _load_compressed(arg.file_obj, file_type=arg.file_type, **kwargs)
        elif arg.file_type in voxel_loaders:
            loaded = voxel_loaders[arg.file_type](
                file_obj=arg.file_obj,
                file_type=arg.file_type,
                resolver=arg.resolver,
                **kwargs,
            )
        else:
            raise NotImplementedError(f"file_type '{arg.file_type}' not supported")

    finally:
        # if we opened the file ourselves from a file name
        # close any opened files even if we crashed out
        if arg.was_opened:
            arg.file_obj.close()

    if not isinstance(loaded, Scene):
        # file name may be used for nodes
        loaded._source = arg
        loaded = Scene(loaded)

    # add on the loading information
    loaded._source = arg
    for g in loaded.geometry.values():
        g._source = arg

    return loaded


def load_mesh(*args, **kwargs) -> Trimesh:
    """
    Load a file into a Trimesh object.

    Parameters
    -----------
    file_obj : str or file object
      File name or file with mesh data
    file_type : str or None
      Which file type, e.g. 'stl'
    kwargs : dict
      Passed to Trimesh constructor

    Returns
    ----------
    mesh
      Loaded geometry data.
    """
    return load_scene(*args, **kwargs).to_mesh()


def _load_compressed(file_obj, file_type=None, resolver=None, mixed=False, **kwargs):
    """
    Given a compressed archive load all the geometry that
    we can from it.

    Parameters
    ----------
    file_obj : open file-like object
      Containing compressed data
    file_type : str
      Type of the archive file
    mixed : bool
      If False, for archives containing both 2D and 3D
      data will only load the 3D data into the Scene.

    Returns
    ----------
    scene : trimesh.Scene
      Geometry loaded in to a Scene object
    """

    # parse the file arguments into clean loadable form
    arg = _parse_file_args(file_obj=file_obj, file_type=file_type, resolver=resolver)

    # store loaded geometries as a list
    geometries = []

    # so loaders can access textures/etc
    archive = util.decompress(file_obj=arg.file_obj, file_type=arg.file_type)
    resolver = resolvers.ZipResolver(archive)

    # try to save the files with meaningful metadata
    # archive_name = arg.file_path or "archive"
    meta_archive = {}

    # populate our available formats
    if mixed:
        available = available_formats()
    else:
        # all types contained in ZIP archive
        contains = {util.split_extension(n).lower() for n in resolver.keys()}
        # if there are no mesh formats available
        if contains.isdisjoint(mesh_formats()):
            available = path_formats()
        else:
            available = mesh_formats()

    for file_name, file_obj in archive.items():
        try:
            # only load formats that we support
            compressed_type = util.split_extension(file_name).lower()

            # if file has metadata type include it
            if compressed_type in ("yaml", "yml"):
                import yaml

                continue
                meta_archive[file_name] = yaml.safe_load(file_obj)
            elif compressed_type == "json":
                import json

                meta_archive[file_name] = json.load(file_obj)
                continue
            elif compressed_type not in available:
                # don't raise an exception, just try the next one
                continue

            # load the individual geometry
            geometries.append(
                load_scene(
                    file_obj=file_obj,
                    file_type=compressed_type,
                    resolver=resolver,
                    **kwargs,
                )
            )

        except BaseException:
            log.debug("failed to load file in zip", exc_info=True)

    # if we opened the file in this function
    # clean up after ourselves
    if arg.was_opened:
        arg.file_obj.close()

    # append meshes or scenes into a single Scene object
    result = append_scenes(geometries)

    # append any archive metadata files
    if isinstance(result, Scene):
        result.metadata.update(meta_archive)

    return result


def load_remote(url: str, **kwargs) -> Scene:
    """
    Load a mesh at a remote URL into a local trimesh object.

    This is a thin wrapper around:
      `trimesh.load_scene(file_obj=url, allow_remote=True, **kwargs)`

    Parameters
    ------------
    url
      URL containing mesh file
    **kwargs
      Passed to `load_scene`

    Returns
    ------------
    loaded : Trimesh, Path, Scene
      Loaded result
    """
    return load_scene(file_obj=url, allow_remote=True, **kwargs)


def _load_kwargs(*args, **kwargs) -> Geometry:
    """
    Load geometry from a properly formatted dict or kwargs
    """

    def handle_scene() -> Scene:
        """
        Load a scene from our kwargs.

        class:      Scene
        geometry:   dict, name: Trimesh kwargs
        graph:      list of dict, kwargs for scene.graph.update
        base_frame: str, base frame of graph
        """
        graph = kwargs.get("graph", None)
        geometry = {k: _load_kwargs(v) for k, v in kwargs["geometry"].items()}

        if graph is not None:
            scene = Scene()
            scene.geometry.update(geometry)
            for k in graph:
                if isinstance(k, dict):
                    scene.graph.update(**k)
                elif util.is_sequence(k) and len(k) == 3:
                    scene.graph.update(k[1], k[0], **k[2])
        else:
            scene = Scene(geometry)

        # camera, if it exists
        camera = kwargs.get("camera")
        if camera:
            scene.camera = camera
            scene.camera_transform = kwargs.get("camera_transform")

        if "base_frame" in kwargs:
            scene.graph.base_frame = kwargs["base_frame"]
        metadata = kwargs.get("metadata")
        if isinstance(metadata, dict):
            scene.metadata.update(kwargs["metadata"])
        elif isinstance(metadata, str):
            # some ways someone might have encoded a string
            # note that these aren't evaluated until we
            # actually call the lambda in the loop
            candidates = [
                lambda: json.loads(metadata),
                lambda: json.loads(metadata.replace("'", '"')),
            ]
            for c in candidates:
                try:
                    scene.metadata.update(c())
                    break
                except BaseException:
                    pass
        elif metadata is not None:
            log.warning("unloadable metadata")

        return scene

    def handle_mesh() -> Trimesh:
        """
        Handle the keyword arguments for a Trimesh object
        """
        # if they've been serialized as a dict
        if isinstance(kwargs["vertices"], dict) or isinstance(kwargs["faces"], dict):
            return Trimesh(**misc.load_dict(kwargs))
        # otherwise just load that puppy
        return Trimesh(**kwargs)

    def handle_export():
        """
        Handle an exported mesh.
        """
        data, file_type = kwargs["data"], kwargs["file_type"]
        if isinstance(data, dict):
            return _load_kwargs(data)
        elif file_type in mesh_loaders:
            return Trimesh(**mesh_loaders[file_type](data, file_type=file_type))

        raise NotImplementedError(f"`{file_type}` is not supported")

    def handle_path():
        from ..path import Path2D, Path3D

        shape = np.shape(kwargs["vertices"])
        if len(shape) < 2:
            return Path2D()
        if shape[1] == 2:
            return Path2D(**kwargs)
        elif shape[1] == 3:
            return Path3D(**kwargs)
        else:
            raise ValueError("Vertices must be 2D or 3D!")

    def handle_pointcloud():
        return PointCloud(**kwargs)

    # if we've been passed a single dict instead of kwargs
    # substitute the dict for kwargs
    if len(kwargs) == 0 and len(args) == 1 and isinstance(args[0], dict):
        kwargs = args[0]

    # (function, tuple of expected keys)
    # order is important
    handlers = (
        (handle_scene, ("geometry",)),
        (handle_mesh, ("vertices", "faces")),
        (handle_path, ("entities", "vertices")),
        (handle_pointcloud, ("vertices",)),
        (handle_export, ("file_type", "data")),
    )

    # filter out keys with a value of None
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    # loop through handler functions and expected key
    for func, expected in handlers:
        if all(i in kwargs for i in expected):
            # all expected kwargs exist
            return func()

    raise ValueError(f"unable to determine type: {kwargs.keys()}")


def _parse_file_args(
    file_obj,
    file_type: Optional[str],
    resolver: Optional[resolvers.ResolverLike] = None,
    allow_remote: bool = False,
    **kwargs,
) -> LoadSource:
    """
    Given a file_obj and a file_type try to magically convert
    arguments to a file-like object and a lowercase string of
    file type.

    Parameters
    -----------
    file_obj : str
      if string represents a file path, returns:
        file_obj:   an 'rb' opened file object of the path
        file_type:  the extension from the file path

     if string is NOT a path, but has JSON-like special characters:
        file_obj:   the same string passed as file_obj
        file_type:  set to 'json'

     if string is a valid-looking URL
        file_obj: an open 'rb' file object with retrieved data
        file_type: from the extension

     if string is none of those:
        raise ValueError as we can't do anything with input

     if file like object:
        ValueError will be raised if file_type is None
        file_obj:  same as input
        file_type: same as input

     if other object: like a shapely.geometry.Polygon, etc:
        file_obj:  same as input
        file_type: if None initially, set to the class name
                    (in lower case), otherwise passed through

    file_type : str
         type of file and handled according to above

    Returns
    -----------
    args
      Populated `_FileArg` message
    """
    # try to save a file path from various inputs
    file_path = None

    # keep track if we opened a file ourselves and thus are
    # responsible for closing it at the end of loading
    was_opened = False

    if util.is_pathlib(file_obj):
        # convert pathlib objects to string
        file_obj = str(file_obj.absolute())

    if util.is_file(file_obj) and file_type is None:
        raise ValueError("`file_type` must be set for file objects!")

    if isinstance(file_obj, str):
        try:
            # clean up file path to an absolute location
            file_path = os.path.abspath(os.path.expanduser(file_obj))
            # check to see if this path exists
            exists = os.path.isfile(file_path)
        except BaseException:
            exists = False
            file_path = None

        # file obj is a string which exists on filesystm
        if exists:
            # if not passed create a resolver to find other files
            if resolver is None:
                resolver = resolvers.FilePathResolver(file_path)
            # save the file name and path to metadata
            # if file_obj is a path that exists use extension as file_type
            if file_type is None:
                file_type = util.split_extension(file_path, special=["tar.gz", "tar.bz2"])
            # actually open the file
            file_obj = open(file_path, "rb")
            # save that we opened it so we can cleanup later
            was_opened = True
        else:
            if "{" in file_obj:
                # if a bracket is in the string it's probably straight JSON
                file_type = "json"
                file_obj = util.wrap_as_stream(file_obj)
            elif "https://" in file_obj or "http://" in file_obj:
                if not allow_remote:
                    raise ValueError("unable to load URL with `allow_remote=False`")

                import urllib

                # remove the url-safe encoding and query params
                file_type = util.split_extension(
                    urllib.parse.unquote(file_obj).split("?", 1)[0].split("/")[-1].strip()
                )
                # create a web resolver to do the fetching and whatnot
                resolver = resolvers.WebResolver(url=file_obj)
                # fetch the base file
                file_obj = util.wrap_as_stream(resolver.get_base())

            elif file_type is None:
                raise ValueError(f"string is not a file: `{file_obj}`")

    if isinstance(file_type, str) and "." in file_type:
        # if someone has passed the whole filename as the file_type
        # use the file extension as the file_type
        path = os.path.abspath(os.path.expanduser(file_type))
        file_type = util.split_extension(file_type)
        if os.path.exists(path):
            file_path = path
            if resolver is None:
                resolver = resolvers.FilePathResolver(file_path)

    # all our stored extensions reference in lower case
    if file_type is not None:
        file_type = file_type.lower()

    # if we still have no resolver try using file_obj name
    if (
        resolver is None
        and hasattr(file_obj, "name")
        and file_obj.name is not None
        and len(file_obj.name) > 0
    ):
        resolver = resolvers.FilePathResolver(file_obj.name)

    return LoadSource(
        file_obj=file_obj,
        file_type=file_type,
        file_path=file_path,
        was_opened=was_opened,
        resolver=resolver,
    )


# loader functions for compressed extensions
compressed_loaders = {
    "zip": _load_compressed,
    "tar.bz2": _load_compressed,
    "tar.gz": _load_compressed,
    "bz2": _load_compressed,
}

# map file_type to loader function
mesh_loaders = {}
mesh_loaders.update(_misc_loaders)
mesh_loaders.update(_stl_loaders)
mesh_loaders.update(_ply_loaders)
mesh_loaders.update(_obj_loaders)
mesh_loaders.update(_off_loaders)
mesh_loaders.update(_collada_loaders)
mesh_loaders.update(_gltf_loaders)
mesh_loaders.update(_xaml_loaders)
mesh_loaders.update(_threedxml_loaders)
mesh_loaders.update(_three_loaders)
mesh_loaders.update(_xyz_loaders)
mesh_loaders.update(_cascade_loaders)

# collect loaders which return voxel types
voxel_loaders = {}
voxel_loaders.update(_binvox_loaders)



try:
    from . import generic as g
except BaseException:
    import generic as g


def generate_primitives():
    primitives = []

    # do it with a flag in case there is more than one ImportError

    e = g.trimesh.primitives.Extrusion()
    e.primitive.polygon = g.trimesh.path.polygons.random_polygon()
    e.primitive.height = 1.0
    primitives.append(e)

    primitives.append(
        g.trimesh.primitives.Extrusion(
            polygon=g.trimesh.path.polygons.random_polygon(), height=293292.322
        )
    )

    primitives.append(e.buffer(0.25))

    primitives.append(
        g.trimesh.primitives.Extrusion(polygon=g.Point([0, 0]).buffer(0.5), height=1.0)
    )

    primitives.append(
        g.trimesh.primitives.Extrusion(polygon=g.Point([0, 0]).buffer(0.5), height=-1.0)
    )

    primitives.append(g.trimesh.primitives.Sphere())
    primitives.append(
        g.trimesh.primitives.Sphere(center=[0, 0, 100], radius=10.0, subdivisions=5)
    )
    primitives.append(g.trimesh.primitives.Box())

    try:
        primitives.append(
            g.trimesh.primitives.Box(center=[102.20, 0, 102.0], extents=[29, 100, 1000])
        )
        raise ValueError("Box shouldn't have accepted `center`!")
    except TypeError:
        # this should have raised a TypeError as `center` is not a kwarg
        pass

    primitives.append(
        g.trimesh.primitives.Box(
            extents=[10, 20, 30],
            transform=g.trimesh.transformations.random_rotation_matrix(),
        )
    )

    primitives.append(g.trimesh.primitives.Cylinder())
    primitives.append(g.trimesh.primitives.Cylinder(radius=10, height=1, sections=40))

    primitives.append(g.trimesh.primitives.Capsule())
    primitives.append(g.trimesh.primitives.Capsule(radius=1.5, height=10))

    return primitives


def test_scaling():
    # try a simple scaling test
    p = g.trimesh.primitives.Sphere(radius=1.0)
    m = p.to_mesh()
    assert g.np.allclose(p.extents, 2.0)
    assert g.np.allclose(m.extents, 2.0)
    p.apply_scale(0.5)
    m.apply_scale(0.5)
    assert g.np.isclose(p.primitive.radius, 0.5)
    assert g.np.allclose(p.extents, 1.0)
    assert g.np.allclose(m.extents, 1.0)
    assert g.np.allclose(p.extents, m.extents, atol=1e-3)

    p = g.trimesh.primitives.Box()
    p.apply_translation([0.5, 0.5, 0.5])
    p.apply_scale(5.0)
    assert g.np.allclose(p.bounds, [[0, 0, 0], [5, 5, 5]])

    try:
        raised = False
        p.apply_scale([2, 3, 4])
    except BaseException:
        raised = True
    if not raised:
        raise ValueError("primitives should raise on non-uniform scaling")

    # now try with more complicated generated data
    prims = [
        g.trimesh.primitives.Sphere(radius=1.0),
        g.trimesh.primitives.Sphere(radius=112.007),
        g.trimesh.primitives.Cylinder(radius=1.0, height=10.0),
        g.trimesh.primitives.Box(),
        g.trimesh.primitives.Box(extents=[12, 32, 31]),
        g.trimesh.primitives.Cylinder(radius=1.1212, height=0.001),
        g.trimesh.primitives.Capsule(radius=1.0, height=7.0),
    ]

    for original in prims:
        perm = [original, original.copy(), original.copy(), original.copy()]
        # try with a simple translation
        perm[1].primitive.transform = g.tf.translation_matrix([0, 0, 7])
        perm[2].apply_transform(g.tf.rotation_matrix(g.np.pi / 4, [0, 0, 1]))
        # try with a gnarly rotation
        perm[3].primitive.transform = g.tf.random_rotation_matrix(translate=1000)

        fields = set(dir(original.primitive))
        ori_radius, ori_height = None, None
        if "radius" in fields:
            ori_radius = original.primitive.radius
        if "height" in fields:
            ori_height = original.primitive.height

        for scale in [1e-2, 0.123, 0.5, 100.2]:
            for po in perm:
                # converting to mesh will do all scaling
                # and transformations on simple discrete
                # copy of primitive and should match with
                # only tessellation differences
                p = po.copy()
                m = p.to_mesh()

                # make sure we have the types we think we do
                assert isinstance(p, g.trimesh.primitives.Primitive)
                assert isinstance(m, g.trimesh.Trimesh)

                assert g.np.allclose(p.extents, m.extents)

                p.apply_scale(scale)
                m.apply_scale(scale)

                # matrix should never have scale
                assert g.tf.is_rigid(p.primitive.transform)

                if ori_radius is not None:
                    assert g.np.isclose(p.primitive.radius, ori_radius * scale)
                if ori_height is not None:
                    assert g.np.isclose(p.primitive.height, ori_height * scale)

                # should be the same size
                assert g.np.allclose(p.extents, m.extents, atol=1e-3 * scale)
                # should be in the same place
                assert g.np.allclose(p.bounds, m.bounds, atol=1e-3 * scale)


def test_mesh_schema():
    # this schema should define every primitive.
    schema = g.trimesh.resources.get_schema("primitive/trimesh.schema.json")
    # make sure a mesh passes the schema
    m = g.trimesh.creation.box()
    g.jsonschema.validate(m.to_dict(), schema)


def test_primitives():
    primitives = generate_primitives()
    kind = {i.__class__.__name__ for i in primitives}
    # make sure our test data has every primitive
    kinds = {"Box", "Capsule", "Cylinder", "Sphere", "Extrusion"}
    assert kind == kinds

    # this schema should define every primitive.
    schema = g.trimesh.resources.get_schema("primitive/primitive.schema.json")

    for primitive in primitives:
        # convert to a dict
        d = primitive.to_dict()
        # validate the output of the to-dict method
        g.jsonschema.validate(d, schema)

        # just triple-check that we have a transform
        # this should have been validated by the schema
        assert g.np.shape(d["transform"]) == (4, 4)
        assert g.trimesh.transformations.is_rigid(d["transform"])
        # make sure the value actually json-dumps
        assert len(g.json.dumps(d)) > 0

        # make sure faces and vertices are correct
        assert g.trimesh.util.is_shape(primitive.faces, (-1, 3))
        assert g.trimesh.util.is_shape(primitive.vertices, (-1, 3))
        # check dtype of faces and vertices
        assert primitive.faces.dtype.kind == "i"
        assert primitive.vertices.dtype.kind == "f"

        assert primitive.volume > 0.0
        assert primitive.area > 0.0

        # convert to base class trimesh
        as_mesh = primitive.to_mesh()

        assert as_mesh.volume > 0.0
        assert as_mesh.area > 0.0

        assert g.np.allclose(primitive.extents, as_mesh.extents)
        assert g.np.allclose(primitive.bounds, as_mesh.bounds)

        assert g.np.isclose(primitive.volume, as_mesh.volume, rtol=0.05)
        assert g.np.isclose(primitive.area, as_mesh.area, rtol=0.05)

        assert primitive.is_winding_consistent
        assert primitive.is_watertight
        assert as_mesh.is_winding_consistent
        assert as_mesh.is_watertight

        # check that overload of dir worked
        assert len([i for i in dir(primitive.primitive) if "_" not in i]) > 0
        if hasattr(primitive, "direction"):
            assert primitive.direction.shape == (3,)

        centroid = primitive.centroid.copy()
        translation = [0, 0, 5]
        primitive.apply_translation(translation)

        # centroid should have translated correctly
        assert g.np.allclose(primitive.centroid - centroid, translation)


def test_sample():
    transform = g.trimesh.transformations.random_rotation_matrix()
    box = g.trimesh.primitives.Box(transform=transform, extents=[20, 10, 100])
    for kwargs in [
        {"step": 8},
        {"step": [10, 0.4, 10]},
        {"count": 8},
        {"count": [10, 3, 5]},
    ]:
        grid = box.sample_grid(**kwargs)
        assert g.trimesh.util.is_shape(grid, (-1, 3))
        assert (box.nearest.signed_distance(grid) > -1e-6).all()


def test_box():
    """
    Test the setter on box primitives
    """
    start = [20, 10, 100]
    box = g.trimesh.primitives.Box(extents=start)
    assert g.np.allclose(box.primitive.extents, start)
    assert g.np.allclose(box.extents, start)
    if g.has_path:
        # check to see if outline function works
        assert g.np.allclose(box.as_outline().extents, start)


def test_transform_return():
    # make sure primitives return `self`
    Box = g.trimesh.primitives.Box

    assert isinstance(Box().apply_transform(g.np.eye(4)), Box)
    assert isinstance(Box().apply_transform(g.transforms[0]), Box)


def test_cyl_buffer():
    # test our inflation of cylinder primitives
    c = g.trimesh.primitives.Cylinder(
        radius=1.0,
        height=10.0,
        transform=g.trimesh.transformations.random_rotation_matrix(),
    )
    # inflate cylinder
    b = c.buffer(1.0)
    assert g.np.isclose(b.primitive.height, 12.0)
    assert g.np.isclose(b.primitive.radius, 2.0)
    # should contain all vertices of source mesh
    assert b.contains(c.vertices).all()
    # should contain line segment
    assert b.contains(c.segment).all()


def test_transform_attribute():
    for primitive in generate_primitives():
        assert hasattr(primitive, "transform")

        assert g.trimesh.util.is_shape(primitive.transform, (4, 4))

        if hasattr(primitive.primitive, "center"):
            assert g.np.allclose(primitive.primitive.center, primitive.transform[:3, 3])


def test_sphere_center():
    s = g.trimesh.primitives.Sphere(center=[0, 0, 100], radius=10.0, subdivisions=5)
    assert g.np.allclose(s.center, [0, 0, 100])

    s.center = [1, 1, 1]
    assert g.np.allclose(s.center, [1, 1, 1])

    s.center[:2] = [0, 3]
    assert g.np.allclose(s.center, [0, 3, 1])


def test_box_bounds_constructor():
    # check to see that we can construct a box using AABB
    bounds = [[0.2, 0.3, 0.4], [10, 11, 12.2]]
    prim = g.trimesh.primitives.Box(bounds=bounds)

    # bounds should match requesta
    assert g.np.allclose(prim.bounds, bounds)

    try:
        # should raise a ValueError
        g.trimesh.primitives.Box(extents=bounds[0], bounds=bounds)
        raise AssertionError("box shouldn't have accepted both bounds and extents!")
    except ValueError:
        pass


def test_copy():
    start = [20, 10, 100]
    # check that copy preserves overridden density and center-mass
    # for both Primitive objects and regular Trimesh objects.
    meshes = [
        g.trimesh.primitives.Box(extents=start),
        g.trimesh.creation.box(extents=start),
    ]

    for box in meshes:
        box.density = 0.3
        box.center_mass = g.np.array([0.1, -0.6, 11.3])
        box.metadata["foo"] = "bar"
        box_copy = box.copy()
        assert box.density == box_copy.density
        assert g.np.allclose(box.center_mass, box_copy.center_mass)
        assert box.metadata["foo"] == box_copy.metadata["foo"]


def test_sphere_subdivisions():
    # make sure we don't subdivide when asked not to
    a = g.trimesh.creation.icosphere(subdivisions=0)
    assert a.faces.shape == g.trimesh.creation.icosahedron().faces.shape

    # now apply the subdivisions ourself
    a = a.subdivide().subdivide()

    # should have the same subdivisions as manually doing this
    b = g.trimesh.primitives.Sphere(subdivisions=2)

    assert a.faces.shape == b.faces.shape


def test_extrude_midplane():
    # check that we extrude along +Z
    e = g.trimesh.primitives.Extrusion(height=10)
    assert g.np.allclose(e.bounds[:, 2], [0.0, 10.0])

    # check that midplane gets half the height correctly
    e = g.trimesh.primitives.Extrusion(height=10, mid_plane=True)
    assert g.np.allclose(e.bounds[:, 2], [-5.0, 5.0])


def test_sphere_subdivisions_radius():
    tf = list(g.random_transforms(4))
    for r in [1e-4, 1.1, 3, 1213.22]:
        for subd in range(4):
            s = g.trimesh.primitives.Sphere(
                radius=r, transform=tf[subd], subdivisions=subd
            )
            rc = g.np.linalg.norm(s.vertices - s.center_mass, axis=1)
            assert g.np.isclose(rc.mean(), r)


if __name__ == "__main__":
    test_extrude_midplane()
